//
//  BFRecordScreenController.swift
//  BFRecordScreenKit_Example
//
//  Created by 胡志强 on 2021/11/23.
//  Copyright © 2021 CocoaPods. All rights reserved.
//

import Foundation
import BFUIKit
import GPUImage
import Photos
import BFCommonKit
import BFFramework

public class BFRecordScreenController: BFBaseViewController {
    
    public var nextActionHandle:(()->Void)?
    public var closeActionHandle:(()->Void)?
    
    // MARK: - 录制参数
    public var asset:PHAsset?
    
    // 视频素材
    public var avasset:AVURLAsset?
    public var recordList:[PQVoiceModel] = [PQVoiceModel]()
    var assetPlayer:AVPlayer?       // 原视频音频播放器
    lazy var recordPlayer:AVAudioPlayer = {// 录音音频播放器
        let player = AVAudioPlayer()
        player.volume = 1
        return player
        
    }()
    var movie :GPUImageMovie?       // 视频预览
    var playView :GPUImageView?     // 视频展示视图
    var isDragingProgressSlder : Bool = false // 是否在拖动进度条
    
    //定义音频的编码参数
    let recordSettings:[String : Any] = [AVSampleRateKey : 44100.0, //声音采样率
                            AVFormatIDKey : kAudioFormatLinearPCM,  //编码格式
                    AVNumberOfChannelsKey : 1,                      //采集音轨
                  AVEncoderBitDepthHintKey: 16,                     // 位深
                 AVEncoderAudioQualityKey : AVAudioQuality.medium.rawValue] //音频质量

    
    // 录音相关
    lazy var recorderManager : BFRecordManager = {
        
        let manager = BFRecordManager(voideModeL: PQVoiceModel())
        manager.cancelRecordHandle = { error in
            
        }
        manager.endRecordHandle = {[weak self] (isTimeout, model) in
            if FileManager.default.fileExists(atPath: model?.wavFilePath ?? ""){
                // 加入到语音数组里
                while let m = self?.recordList.last{
                    if m.startTime < model!.startTime {
                        self?.recordList.removeLast()
                    }else if m.endTime < model!.startTime {
                        m.endTime = model!.startTime
                    }else{
                        break
                    }
                }
                self?.recordList.append(model!)
                self?.drewRecordProgessLable()
            }
            
        }
        return manager
    }()

    // MARK: - 视图参数
    var beginOnStartBtn:Bool = false
    var touchStart:CGPoint = CGPoint(x: 0, y: 0)
    var avplayerTimeObserver: NSKeyValueObservation?

    lazy var playBtn:UIButton = {
        let btn = UIButton(frame: view.bounds)
        btn.addTarget(self, action: #selector(playVideo(btn:)), for: .touchUpInside)
        return btn
    }()
    
    lazy var bottomeView:UIImageView = {
        let iv = UIImageView(image: imageInRecordScreenKit(by: "bottom_shadow"))
        iv.contentMode = .scaleAspectFill
        iv.isUserInteractionEnabled = true
        // 拖曳手勢
        let pan = UIPanGestureRecognizer(target: self, action: #selector(pan(recognizer:)))
        pan.delegate = self
        pan.minimumNumberOfTouches = 1
        pan.maximumNumberOfTouches = 1
        iv.addGestureRecognizer(pan)
        
        return iv
    }()
    
    lazy var recordBtn:UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "mic1"), for: .normal)
        btn.adjustsImageWhenHighlighted = false
        btn.addTarget(self, action: #selector(startRecord(btn:)), for: .touchDown)
        btn.addTarget(self, action: #selector(endRecord(btn:)), for: .touchUpInside)
        return btn
    }()
    
    lazy var progessSilde:BFPlayerSlider = {
        let sliderView = BFPlayerSlider()
        let tjbV = UIView(frame: CGRect(x: 0, y: 0, width: 4, height: 16))
        tjbV.backgroundColor = .white
        let thbImage = tjbV.graphicsGetImage()//UIImage(named: "icon_point")
        sliderView.setMinimumTrackImage(thbImage, for: .normal)
        sliderView.setMaximumTrackImage(thbImage, for: .normal)
        sliderView.setThumbImage(thbImage, for: .highlighted)
        sliderView.setThumbImage(thbImage, for: .normal)
        sliderView.maximumTrackTintColor = UIColor.hexColor(hexadecimal: "#303030")
        sliderView.minimumTrackTintColor = UIColor.hexColor(hexadecimal: "#303030")
//        sliderView.minimumTrackTintColor = UIColor.hexColor(hexadecimal: "#FA6400")
        sliderView.addTarget(self, action: #selector(sliderTouchBegan(sender:)), for: .touchDown)
        sliderView.addTarget(self, action: #selector(sliderTouchEnded(sender:)), for: .touchUpInside)
        sliderView.addTarget(self, action: #selector(sliderTouchEnded(sender:)), for: .touchUpOutside)
        sliderView.addTarget(self, action: #selector(sliderTouchEnded(sender:)), for: .touchCancel)
        return sliderView
    }()
    
    lazy var closeBtn:UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "xx"), for: .normal)
        btn.addTarget(self, action: #selector(close), for: .touchUpInside)
        return btn
    }()
    
    lazy var nextBtn:UIButton = {
        let btn = UIButton(type: .custom)
        btn.setImage(imageInRecordScreenKit(by: "gou"), for: .normal)
        btn.addTarget(self, action: #selector(nextAction), for: .touchUpInside)
        return btn
    }()
    
    lazy var toolV : BFIntroduceToolView = {
        let toolV = BFIntroduceToolView()
        toolV.centerY = view.centerY
        
        toolV.choosedToolHandle = {[weak self] tool in
            guard let sself = self else {
                return
            }
            tool.center = sself.view.center
            sself.view.addSubview(tool)
        }
        
        return toolV
        
    }()
    //MARK: ------------------ 生命周期
    deinit {
        cleanMovieTarget()
        NotificationCenter.default.removeObserver(self)
        avplayerTimeObserver?.invalidate()
        recorderManager.stopRecord(isCancel: true)
        assetPlayer?.pause()
        recordPlayer.pause()
        
    }
    
    public override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        self.navigationController?.isNavigationBarHidden = true
        hiddenNavigation()
    }
    
    public  override func viewDidLoad(){
        super.viewDidLoad()
        _ = disablePopGesture()
        
        view.backgroundColor = .lightGray
        
        playView = GPUImageView(frame: view.bounds)
        view.addSubview(playView!)
        fetchVideo()
        
        view.addSubview(playBtn)
        view.addSubview(bottomeView)
        view.addSubview(toolV)
        bottomeView.addSubview(recordBtn)
        bottomeView.addSubview(progessSilde)
        bottomeView.addSubview(closeBtn)
        bottomeView.addSubview(nextBtn)
        
        if checkStatus() {
            try? AVAudioSession.sharedInstance().setCategory(.playAndRecord, options: .defaultToSpeaker)
        }
        
    }
    
    public override func viewWillLayoutSubviews() {
        super.viewWillLayoutSubviews()
        bottomeView.snp.makeConstraints { make in
            make.left.bottom.right.equalToSuperview()
            make.height.equalTo(adapterWidth(width: 220))
        }
        
        recordBtn.snp.makeConstraints { make in
            make.width.height.equalTo(120)
            make.centerX.equalToSuperview()
            make.top.equalTo(27)
        }
        
        closeBtn.snp.makeConstraints { make in
            make.left.equalToSuperview()
            make.width.height.equalTo(60)
            make.top.equalTo(recordBtn.snp.bottom).offset(-10)
        }
        nextBtn.snp.makeConstraints { make in
            make.right.equalToSuperview()
            make.top.width.height.equalTo(closeBtn)
        }
        
        progessSilde.snp.makeConstraints { make in
            make.left.equalTo(closeBtn.snp.right).offset(16)
            make.right.equalTo(nextBtn.snp.left).offset(-16)
            make.centerY.equalTo(closeBtn)
            make.height.equalTo(20)
        }
        
    }
    
    // MARK: - 按钮事件响应
    
    // 觸發拖曳手勢後 執行的動作
    @objc func pan(recognizer: UIPanGestureRecognizer) {
        // 設置 UIView 新的位置
        if !checkStatus(show: false) {
            return
        }

        let point = recognizer.location(in: bottomeView)

        switch recognizer.state {
        case .began:
            touchStart = recognizer.location(in: bottomeView)
            beginOnStartBtn = recordBtn.frame.contains(touchStart)
        case .changed:
            if beginOnStartBtn == true {
                let nowPoint = recognizer.location(in: bottomeView)
                BFLog(1, message: "nowPoint x: \(nowPoint.x) y:\(nowPoint.y)")
                if recordBtn.frame.contains(nowPoint) {
                    
                } else {
                    
                }
            }

        case .ended:
            BFLog(1, message: "移动结束 \(beginOnStartBtn)")
            if beginOnStartBtn == true {
                beginOnStartBtn = false
                if recordBtn.frame.contains(point) {
                    // 结束录制
                    endRecord(btn: recordBtn)
                } else {
                    cancleRecord()
                }
            }

        default:
            break
        }
    }
    
    @objc func startRecord(btn:UIButton){
        btn.setImage(imageInRecordScreenKit(by: "mic2"), for: .normal)
        BFLog(1, message: "start \(UIControl.Event.touchDown)")
        let model = PQVoiceModel()
        model.startTime = CMTimeGetSeconds(assetPlayer?.currentItem?.currentTime() ?? CMTime.zero)
        recorderManager.voiceModel? = model
        recorderManager.startRecord(index: recordList.count)
        movie?.startProcessing()
        assetPlayer?.volume = 0
        assetPlayer?.play()
    }
    
    @objc func endRecord(btn:UIButton){
        recordBtn.setImage(imageInRecordScreenKit(by: "mic1"), for: .normal)
        // 存储录音
        recorderManager.endRecord()
        pause()
    }
    
    func cancleRecord(){
        recordBtn.setImage(imageInRecordScreenKit(by: "mic1"), for: .normal)
        recorderManager.cancleRecord()
        pause()
    }
    @objc func close(){
        pause()
        closeActionHandle?()
    }
    
    @objc func nextAction(){
        nextActionHandle?()
    }
    
    @objc func playVideo(btn:UIButton){
        btn.isSelected = !btn.isSelected
        BFLog(1, message: "点击")
        if btn.isSelected{
            play()

        }else{
            pause()
        }
    }

    @objc public func sliderTouchBegan(sender _: UISlider) {
//        isDragingProgressSlder = true
//        pause()
    }

    @objc public func sliderTouchEnded(sender: UISlider) {
//        changeProgress(progress: sender.value)
//        isDragingProgressSlder = false
//        play()
    }
    
    // MARK: - 权限申请
    func checkStatus(show: Bool = true) -> Bool {
        let status = AVCaptureDevice.authorizationStatus(for: .audio)
        switch status {
        case .denied, .restricted:

            BFLog(message: "麦克风权限被拒绝，请在设置中打开")
            if show {
                let remindData = BFBaseModel()
                remindData.title = "票圈视频需要访问你的麦克风才能录音"
                remindData.summary = ""
                let remindView = BFRemindView(frame: CGRect(x: 0, y: 0, width: cScreenWidth, height: cScreenHeigth))
                remindView.isBanned = true
                remindView.confirmBtn.setTitle("去设置", for: .normal)
                UIApplication.shared.keyWindow?.addSubview(remindView)
                remindView.remindData = remindData
                remindView.remindBlock = { item, _ in
                    if item.tag == 2 {
                        openAppSetting()
                    }
                }
            }
            return false
        case .authorized:
            return true
        case .notDetermined:
            if show {
                requestMicroPhoneAuth()
            }
            return false
        default:
            break
        }

        return false
    }
    
    func requestMicroPhoneAuth() {
        AVCaptureDevice.requestAccess(for: .audio) { granted in
            if !granted {
                BFLog(message: "麦克风权限被拒绝，请在设置中打开")
            }
        }
    }
    
    // MARK: - 音视频处理
    func playRecord(at duration:CMTime){
        if recordList.count > 0, let player = try? AVAudioPlayer(contentsOf: URL(fileURLWithPath: (recordList.first?.wavFilePath)!)) {
            self.recordPlayer = player
            self.recordPlayer.volume = 1
            self.recordPlayer.play()
        }else{
            self.recordPlayer.pause()
        }
    }
    
    func play(){
        cShowHUB(superView: nil, msg: "开始播放")
        assetPlayer?.volume = 0.5
        movie?.startProcessing()
        let second = assetPlayer?.currentItem?.currentTime()
        
        assetPlayer?.play()
        assetPlayer?.seek(to: second ?? CMTime.zero)
        playRecord(at: CMTime.zero)
    }
    
    func pause(){
        cShowHUB(superView: nil, msg: "暂停播放")
        movie?.cancelProcessing()
        assetPlayer?.pause()
        recordPlayer.pause()
    }
    
    func fetchVideo(){
        if let asset = self.asset {
            let options = PHVideoRequestOptions()
            options.isNetworkAccessAllowed = true
            options.deliveryMode = .automatic
            
            PHImageManager.default().requestPlayerItem(forVideo:asset, options: options, resultHandler: { [weak self] playerItem, info in

                guard let item = playerItem else {
                    cShowHUB(superView: nil, msg: "视频获取失败")
                    return
                }
                self?.setAudioPlay(item: item)
                self?.setVideoPlay(item: item)
                
            })
            
            let option = PHImageRequestOptions()
            option.isNetworkAccessAllowed = true //允许下载iCloud的图片
            option.resizeMode = .fast
            option.deliveryMode = .highQualityFormat
            PHImageManager.default().requestImage(for: asset,
                                           targetSize: self.view.bounds.size,
                                          contentMode: .aspectFit,
                                              options: option)
            { (image, nil) in
                 //image就是图片
                if image != nil {
                    let pic = GPUImagePicture(image: image)
                    let filet = GPUImageFilter()
                    pic?.addTarget(filet)
                    filet.addTarget(self.playView)
                    pic?.processImage()
                }
                
            }
            
//          使用copy资源到本地的方式
//            let outFilePath = NSHomeDirectory().appending("/Documents/simple.mp4")
//            let outFileUrl = URL(fileURLWithPath: outFilePath)
//            if FileManager.default.fileExists(atPath: outFilePath) {
//                try? FileManager.default.removeItem(atPath: outFilePath)
//            }
//
//            let assetResources = PHAssetResource.assetResources(for: asset)
            
//            if let rsc = assetResources.first(where: { res in
//                res.type == .video  || res.type == .pairedVideo
//            }) {
//                PHAssetResourceManager.default().writeData(for: rsc, toFile:outFileUrl, options: nil) {[weak self] error in
//                    if error == nil {
//                        DispatchQueue.main.async {[weak self] in
//                            self?.setVideoPlay(url: outFileUrl)
//                            self?.setAudioPlay(url: outFileUrl)
//                        }
//                    }else{
//                        BFLog(1, message: "导出视频相exportAsynchro faile")
//                    }
//                }
//            }else{
//
//            }
            PHCachingImageManager().requestAVAsset(forVideo: asset, options: nil, resultHandler: {[weak self] (asset: AVAsset?, audioMix: AVAudioMix?, info) in
                if let urlass = asset as? AVURLAsset {
                    self?.avasset = urlass
                }
            })
        }

    }
    
    func setVideoPlay(item:AVPlayerItem){
        if movie != nil {
            cleanMovieTarget()
        }
        movie = GPUImageMovie(playerItem: item)
//        movie?.runBenchmark = true
        movie?.playAtActualSpeed = true

        let filter = GPUImageFilter()
        movie?.addTarget(filter)
        filter.addTarget(playView)
    }
    
    func setAudioPlay(item:AVPlayerItem){

        if let playItem = assetPlayer?.currentItem {
            NotificationCenter.default.removeObserver(self, name: .AVPlayerItemDidPlayToEndTime, object: playItem)
            assetPlayer?.replaceCurrentItem(with: item)
        }else {
            assetPlayer = AVPlayer(playerItem: item)
            avplayerTimeObserver = assetPlayer?.addPeriodicTimeObserver(forInterval: CMTime(value: 1, timescale: 10), queue: DispatchQueue.global()) {[weak self] time in
             //    进度监控
                BFLog(1, message: "curr:\(CMTimeGetSeconds(time))")
                if CMTimeGetSeconds(item.duration) > 0, !(self?.isDragingProgressSlder ?? false) {
                    DispatchQueue.main.async { [weak self] in
                        self?.progessSilde.value = Float(CMTimeGetSeconds(time) / CMTimeGetSeconds(item.duration))
                    }
                }
            } as? NSKeyValueObservation
        }

        
        NotificationCenter.default.addObserver(forName: .AVPlayerItemDidPlayToEndTime, object: assetPlayer?.currentItem, queue: .main) { [weak self] notify in
            BFLog(message: "AVPlayerItemDidPlayToEndTime = \(notify)")
            self?.assetPlayer?.seek(to: CMTime.zero)
            self?.playBtn.isSelected = false
        }
    }

    
    func cleanMovieTarget(){
        movie?.cancelProcessing()
        movie?.targets().forEach({ target in
            if let objc = target as? GPUImageOutput{
                objc.removeAllTargets()
            }
        })
        movie?.removeAllTargets()
    }
    
    //MARK: - 录音对应图像绘制
    
    func changeProgress(progress:Float) {
        if let item = assetPlayer?.currentItem {
            let duration = CMTimeGetSeconds(item.duration)
            item.seek(to: CMTime(value: CMTimeValue(progress * Float(duration) * 100), timescale: 100)) { finished in
                if finished{
                    BFLog(1, message: "拖动成功")
                }
            }
        }
    }
    
    func drewRecordProgessLable(){
        
    }
}

extension BFRecordScreenController:GPUImageMovieDelegate {
    public func didCompletePlayingMovie(){
        BFLog(1, message: "播放结束")
        
    }
}

extension BFRecordScreenController:AVAudioRecorderDelegate {
    
}
